# Stanford::WebApp::Logdump - Web interface to create/edit/delete logdumps
#
# Written by Jon Robertson <jonrober@stanford.edu>
# Copyright 2011-2012 Board of Trustees, Leland Stanford Jr. University

#############################################################################
# Modules and declarations
#############################################################################

## no critic (Modules::RequireNoMatchVarsWithUseEnglish);
## no critic (CodeLayout::ProhibitParensWithBuiltins);

package Stanford::WebApp::Logdump;

use base qw (CGI::Application);
use CGI::Application::Plugin::AutoRunmode;
use CGI::Application::Plugin::ConfigAuto qw(cfg);
use CGI::Application::Plugin::DBH qw(dbh_config dbh);
use CGI::Application::Plugin::Forward;
use CGI::Application::Plugin::MessageStack;
use CGI::Application::Plugin::Session;
use CGI::Application::Plugin::TT;

use strict;
use warnings;
use autodie;

use AppConfig;
use Carp;
use English;
use HTML::FormFu;
use Net::Remctl;
use POSIX qw(strftime);

use Stanford::Schema::Logdumps;
use Stanford::WebApp::Logdump::Util qw(run_remctl);

our $LSDB = 'lsdb.stanford.edu';

my $DEBUG = 1;
#############################################################################
# Setup functions
#############################################################################

sub cgiapp_init {
    my ($self) = @_;
    ## my $url = '/cgi-bin/logdump-request';
    my $url = $ENV{'SCRIPT_FILENAME'};
    $url =~ s{/usr/share/logdump-request/cgi-bin}{}xsm;
    $self->param('url', $url);

    # Configure the template
    $self->tt_config(
        TEMPLATE_OPTIONS => {
            INCLUDE_PATH => [
                $self->param('rootdir'),
                '/usr/share/stanford-web-template/html',
            ],
            WRAPPER  => $self->param('wrapperdir') . '/wrapper.tt2',
            ABSOLUTE => 1,
        },
    );

    # Generate the logdump Schema.
    my $schema = Stanford::Schema::Logdumps->connect(
        $self->param('dsn'),
        $self->param('db_user'),
        $self->param('db_pass')
    );
    $self->param('schema', $schema);

    # We don't want any messages to persist from screen to screen.
    $self->capms_config(-automatic_clearing => 1);

    return;
}

# Prerunning the application, make sure rm works if you're using GET or POST.
# http://www.perlmonks.org/?node_id=748939
sub cgiapp_prerun {
    my ($self) = @_;
    if (!defined $self->query->param('rm')) {
        $self->prerun_mode($self->query->url_param('rm'));
    }

    return;
}

# Create and return a new HTML::FormFu object with some default values.
sub create_formfu {
    my ($self) = @_;

    my $url  = $self->param('url');
    my $form = HTML::FormFu->new(
        {
            action => $url,
            method => 'POST'
        }
    );
    $form->config_file_path($self->param('rootdir'));
    $self->param('formfu', $form);
    return $form;
}

#############################################################################
# Misc routines
#############################################################################

# Given the record a user has been searching for, validate their access.
# If they have access to the record, return 1.  If they do not, set an error
# message and return 0.  If the record does not exist, set an error message
# and return 0.
sub validate_access {
    my ($self, $resultset) = @_;
    if (!defined $resultset || !$resultset) {
        my $msg = 'There was a problem loading the logdump record. Please '
          . 'try again, or file a HelpSU ticket if the problem persists.';
        $self->push_message(
            -message        => $msg,
            -classification => 'error'
        );
        return 0;
    } elsif (!$self->user_has_access($resultset)) {
        my $msg = 'You do not have access to edit the given logdump.';
        $self->push_message(
            -message        => $msg,
            -classification => 'error'
        );
        return 0;
    }

    return 1;
}

# Given a record, see if the user has access to it.  This is used to validate
# against the user trying to edit or view random ids.
sub user_has_access {
    my ($self, $resultset) = @_;
    return 1 if $resultset->requester eq $ENV{REMOTE_USER};
    return 0;
}

# Checks for the existance of the directories to drop files (logreports and
# logdumps) into.  If they do not exist, create them.
sub setup_destination {
    my ($self, $dir) = @_;

    progress('entering setup_destination');
    progress("directory: $dir");

    my ($stdout, $stderr, $rc);

    # Set up all paths under the main path, until we reach the root group,
    # dept, user, or class directory.  Create missing directories and make
    # sure all directories have the correct ownership.
    my $backdir = $dir;
    do {
        progress("ENV{'KRB5CCNAME'} is $ENV{'KRB5CCNAME'}");
        progress("remctl $LSDB afsdir chkexist $backdir");
        ($stdout, $stderr, $rc) =
          run_remctl($LSDB, 0, q{}, 'afsdir', 'chkexist', $backdir);
        progress("remctl returned '($stdout, $stderr, $rc)'");

        if ($rc != 0) {
            progress("remctl $LSDB afsdir chkexist $backdir result is non-zero");
            progress("will try remctl afsdir mkdir $backdir");
            ($stdout, $stderr, $rc) =
              remctl($LSDB, 0, q{}, 'afsdir', 'mkdir', $backdir);
            progress("remctl returned '($stdout, $stderr, $rc)'");
            if ($rc != 0) {
                return 0;
            }
        }
        progress("remctl $LSDB afsdir setacl $backdir service.www-logs read");
        ($stdout, $stderr, $rc) =
          remctl($LSDB, 0, q{}, 'afsdir', 'setacl', $backdir,
                'service.www-logs', 'read');
        progress("remctl returned '($stdout, $stderr, $rc)'");

        $backdir =~ s{/[^/]+/?$}{/}xsm;
    } until $backdir =~ m{^/afs/\.?ir(\.stanford\.edu)?/(group|dept|users/./.|class|class/archive/\w+/[\w-]+/[\w-]+\.\d+)/[^/]+/$}xsm;

    # Create any of the requested dirs that do not exist.
    my $path = $dir . 'logdumps';

    my $result = remctl($LSDB, 0, q{}, 'afsdir', 'chkexist', $path);
    if ($result->status != 0) {
        $result = remctl($LSDB, 0, q{}, 'afsdir', 'mkdir', $path);
        if ($result->status != 0) {
            return 0;
        }
    }

    $result = remctl($LSDB, 0, q{}, 'afsdir', 'setacl', $path,
                      'service.www-logs', 'all');
    if ($result->status != 0) {
        return 0;
    }

    return 1;
}

sub progress {
    my ($msg) = @_;
    if ($DEBUG) {
        carp "DEBUG: $msg";
    }
    return;
}

#############################################################################
# Views
#############################################################################

#############################################################################
# Runmodes
#############################################################################

# The main screen, a list of all current requests.
sub index : StartRunmode {
    my ($self) = @_;

    my $schema = $self->param('schema');
    my %search = (
        'requester' => $ENV{REMOTE_USER},
        'inactive'  => 0
    );
    my %order = ('order_by' => 'name,destination');

    my @requests = $schema->resultset('RequestSet')->search(
        \%search,
        \%order
    );

    my %params = (list => \@requests);

    my $template = $self->tt_process('index.tt2', \%params);
    return $template;
}

# The first page for a new request, asking to be given the name and type.
sub new_first : Runmode {
    my ($self) = @_;

    # Return to the main screen on any cancel.
    if ($self->query->param('cancel')) {
        return $self->forward('index');
    }
    my $schema = $self->param('schema');

    my $formfu = $self->create_formfu;
    $formfu->load_config_file('new_first.yml');
    $formfu->process($self->query);
    my $form = $formfu->render;

    # Create filler record and forward to the proper sub-page if we're valid.
    if ($formfu->submitted_and_valid) {
        my %record_new = (
            'requester'        => $ENV{REMOTE_USER},
            'type'             => scalar($self->query->param('type')),
            'name'             => scalar($self->query->param('name')),
            'compression_type' => 'bzip2',
            'destination'      => q{},
        );
        my $logdump = $schema->resultset('RequestSet')->create(\%record_new);
        $self->param('id', $logdump->id);
        return $self->forward('edit');
    }

    # Otherwise our first visit, or a visit with errors.
    my %params = (form => $form);
    $params{'url'} = $self->param('url');
    my $template = $self->tt_process('new_first.tt2', \%params);
    return $template;
}

# Provide and edit the edit form.
sub edit : Runmode {
    my ($self) = @_;

    progress('entering edit : Runmode');

    my $id       = $self->query->param('id') || $self->param('id');
    my $schema   = $self->param('schema');
    my $logdump  = $schema->resultset('RequestSet')->find($id);
    my $formtype = 'edit_' . $logdump->type . '.yml';

    if (! $self->validate_access($logdump)) {
        return $self->forward('index');
    }

    # Use the HTML::FormFu object for validation.
    my $form = $self->create_formfu;
    $form->load_config_file($formtype);
    $form->auto_constraint_class('constraint_%t');
    $form->process($self->query);

    # Save the record.
    if ($form->submitted_and_valid) {
        $form->model->update($logdump);

        # Harder part is saving/updating paths.  Delete any existing, then
        # add the current request.
        my $paths = $schema->resultset('Path')
            ->search ({'request_set' => $id})->delete;
        if ($logdump->type eq 'group') {
            my %record_new = ('request_set' => $id,
                          'uri'         => $self->query->param('request'),
                          'action'      => 'include',
                );
            $schema->resultset('Path')->create(\%record_new);
        } elsif ($logdump->type eq 'user') {
            my %record_new = (
                'request_set' => $id,
                'uri'         => q{/~} . $ENV{REMOTE_USER},
                'action'      => 'include',
                );
            $schema->resultset('Path')->create(\%record_new);
        } elsif ($logdump->type eq 'proxy') {
            foreach my $proxy (split (m{[\r\n]+}xsm,
                                      $self->query->param('requests'))) {
                $proxy =~ s{^https?://}{}xsm;
                $proxy =~ s{^([^.]+)$}{$1.stanford.edu}xsm;
                my %record_new = ('request_set' => $id,
                              'uri'         => $proxy,
                              'action'      => 'include',
                    );
                $schema->resultset('Path')->create (\%record_new);
            }
        } elsif ($logdump->type eq 'compound') {
            foreach my $proxy (split (m{[\r\n]+}xsm,
                                      $self->query->param('requests_include'))) {
                my %record_new = ('request_set' => $id,
                              'uri'         => $proxy,
                              'action'      => 'include',
                    );
                $schema->resultset('Path')->create(\%record_new);
            }
            foreach my $proxy (split (m{[\r\n]+}xsm,
                                      $self->query->param('requests_exclude'))) {
                my %record_new = ('request_set' => $id,
                              'uri'         => $proxy,
                              'action'      => 'exclude',
                    );
                $schema->resultset('Path')->create(\%record_new);
            }
        }

        # Create/set permissions on the destination directory.
        if ($self->setup_destination (scalar($self->query->param('destination')))) {
            my $msg = 'Successfully updated record';
            $self->push_message (-message        => $msg,
                                 -classification => 'message');
            return $self->forward('index');
        } else {
            my $msg = 'There was a problem creating and setting the correct '
                .'permissions on the destination.  Please check the path '
                .'given and try again.';
            $self->push_message (-message        => $msg,
                                 -classification => 'error');
        }
    } else {
        $form->model->default_values ($logdump);
        my (@paths) = $schema->resultset('Path')
            ->search ({'request_set' => $id});

        if ($logdump->type eq 'group') {
            my ($first) = @paths;
            my $edit = $form->get_element ({name => 'request'});
            if (defined $first) {
                $edit->value($first->uri);
            }
        } elsif ($logdump->type eq 'proxy') {
            my $edit = $form->get_element ({name => 'requests'});
            my $value = q{};
            foreach my $path (@paths) {
                $value .= $path->uri . "\n";
            }
            $edit->value ($value);
        } elsif ($logdump->type eq 'compound') {
            my $include = $form->get_element ({name => 'requests_include'});
            my $exclude = $form->get_element ({name => 'requests_exclude'});
            my $value_inc = q{};
            my $value_exc = q{};
            foreach my $path (@paths) {
                if ($path->action eq 'include') {
                    $value_inc .= $path->uri . "\n";
                } else {
                    $value_exc .= $path->uri . "\n";
                }
            }
            $include->value ($value_inc);
            $exclude->value ($value_exc);
        }
    }

    my %params;
    $params{'form'} = $form;
    $params{'url'}  = $self->param('url');
    my $template = $self->tt_process('edit.tt2', \%params);

    progress('exiting edit : Runmode');

    return $template;
}

# Confirm mode for a request to delete a logdump.  No actual screen, only
# perform the action, push a message, and forward back to the index.
sub purge_confirm : Runmode {
    my ($self) = @_;

    my ($msg, $type);
    my $id = $self->query->param('id');
    my $schema = $self->param('schema');
    my $logdump = $schema->resultset('RequestSet')->find($id);

    if (! $self->validate_access($logdump)) {
        return $self->forward('index');
    }

    $logdump->inactive (1);
    if ($logdump->update) {
        $msg = 'The logdump was successfully deleted.';
        $type = 'message';
    } else {
        $msg = 'There was an error deleting this logdump.  Please try '
            .'again, or file a HelpSU ticket if the problem persists.';
        $type = 'error';
    }
    $self->push_message (
                         -message        => $msg,
                         -classification => $type,
                        );

    return $self->forward('index');
}

# Display the page prompting for confirmation after requesting to delete a
# logdump.
sub purge : Runmode {
    my ($self) = @_;

    my $id = $self->query->param('id');
    my $schema = $self->param('schema');
    my $logdump = $schema->resultset('RequestSet')->find($id);
    if (! $self->validate_access($logdump)) {
        return $self->forward('index');
    }

    my %params = (id => $id);
    $params{'url'} = $self->param('url');
    my $template = $self->tt_process('purge.tt2', \%params);
    return $template;
}

# A request to touch the last updated field in the database in order to
# verify that a logdump is still in use.
sub renew : Runmode {
    my ($self) = @_;

    my $id = $self->query->param('id');
    my $schema = $self->param('schema');
    my $logdump = $schema->resultset('RequestSet')->find($id);
    if (! $self->validate_access($logdump)) {
        return $self->forward('index');
    }

    my $now = strftime('%Y-%m-%d %T', localtime(time));
    eval {
        $logdump->rs_updated($now);
        $logdump->update;
    };
    my ($msg, $type);
    if ($EVAL_ERROR) {
        $msg  = 'There was an error renewing this logdump.  Please try '
          . 'again, or file a HelpSU ticket if the problem persists.';
        $type = 'error';
    } else {
        $msg = 'Your logdump has been renewed.  You should not be asked to '
            . 'verify again for eleven months.';
        $type = 'message';
    }
    $self->push_message (
                         -message        => $msg,
                         -classification => $type,
                        );

    return $self->forward('index');
}

1;

##############################################################################
# Documentation
##############################################################################

=head1 NAME

name - description

=head1 SYNOPSIS

name [B<--help>]

=head1 DESCRIPTION

=head1 OPTIONS

B<--help>

Prints the perldoc information (this document) for the script.

=head1 AUTHORS

Jon Robertson <jonrober@stanford.edu>

=cut
